from flask import Flask, Response

from functools import wraps

from http import HTTPStatus

import json


class API:

    def __init__(self, host: str, port: int, debug: bool):
        self._host = host
        self._port = port
        self._debug = debug

        self._flask: Flask = Flask(__name__)
        self.health = self.route('/health')(self.health)

    def health(self):
        return {'status': 'healthy'}

    def route(self, rule, **kwargs):
        def decorator(fn):
            @wraps(fn)
            def internal():
                try:
                    data = fn()
                    return Response(json.dumps(data), content_type='application/json')
                except Exception as e:
                    return Response(
                        json.dumps({'error': str(e)}),
                        status=HTTPStatus.INTERNAL_SERVER_ERROR,
                        content_type='application/json'
                    )

            return self._flask.route(rule, **kwargs)(internal)
        return decorator

    def start(self):
        self._flask.run(host=self._host, port=self._port, debug=self._debug)



from subprocess import call

from typing import List

import os


def run(cmd_args: List[str], silent=True) -> None:
    """
    Run a shell command with it's arguments.
    Args:
        cmd_args:  The command & arguments.
        silent: Whether to swallow the output.
    """

    if silent:
        with open(os.devnull, 'w') as fnull:
            status = call(cmd_args, stdout=fnull)
    else:
        status = call(cmd_args)

    if status != 0:
        # TODO: Retrieve cmd stdout/stderr
        raise OSError(f"an error ({status})  occured while running command {cmd_args[0]}")



from flask import request

from functools import wraps


def authenticated(password):
    def decorator(fn):
        @wraps(fn)
        def internal(*args, **kwargs):
            expected = request.headers.get('Authorization', 'secret')
            if expected == password:
                return fn(*args, **kwargs)
            else:
                raise ValueError('forbidden')

        return internal

    return decorator


from dalloriam.api import API

from unittest import mock


def test_api_initializes_properly():
    a = API('host', 4242, True)
    assert a._host == 'host'
    assert a._port == 4242
    assert a._debug is True


from dalloriam import filesystem, shell


def build(build_path: str, image_name: str, tag: str = 'latest') -> None:
    """
    Builds a docker image from a directory.
    Args:
        build_path: Path that contains the image-related files.
        image_name: Desired name of the built image.
        tag:        Desired tag of the built image.
    """
    # TODO: Give options to log or not.

    with filesystem.location(build_path):
        image_name = f'{image_name}:{tag}'
        shell.run([
            'docker',
            'build',
            '-t',
            image_name,
            '.'
        ], silent=True)


def push(image_name: str, tag: str = 'latest'):
    image_name = f'{image_name}:{tag}'
    shell.run([
        'docker',
        'push',
        image_name
    ], silent=False)



from dalloriam.shell import runner

from typing import List

from unittest import mock

import pytest


@pytest.mark.parametrize(
    'silent,arguments,want_stdout,status,want_err',
    [
        (True, ['docker', 'run', 'stuff'], True, 0, False),
        (False, ['docker', 'run', 'stuff'], False, 0, False),
        (False, ['docker', 'run', 'stuff'], False, 1, True),
    ]
)
def test_run(silent: bool, arguments: List[str], want_stdout: bool, status: int, want_err: bool):
    try:
        with mock.patch.object(runner, 'call', return_value=status) as mock_call:
            runner.run(arguments, silent)
            mock_call.assert_called_once()
            assert isinstance(mock_call, mock.MagicMock)

            assert mock_call.call_args[0][0] == arguments

            if want_err:
                pytest.fail(f'expected an exception for status {status}.')  # pragma: nocover

    except OSError:
        if not want_err:
            pytest.fail(f'did not expect exception for status {status}.')  # pragma: nocover



from dalloriam.api import authentication

from tests.mocks.request import MockRequest

from unittest import mock

import pytest


def test_authorization_defaults_to_secret_if_no_key_provided():
    with mock.patch.object(authentication, 'request', MockRequest(headers={})):
        m = mock.MagicMock()
        authentication.authenticated('secret')(m)()
        m.assert_called_once()


def test_authorization_doesnt_call_internal_when_password_is_incorrect():
    with mock.patch.object(authentication, 'request', MockRequest(headers={'Authorization': 'Bad'})):
        m = mock.MagicMock()
        try:
            authentication.authenticated('good_password')(m)()
            pytest.raises('authenticated did not raise an exception on bad password!')  # pragma: nocover
        except ValueError:
            pass



from dalloriam.docker import builder

from tests.mocks.location import mock_location

from unittest import mock


def test_builder_build_calls_shell_run_properly():
    with mock.patch.object(builder.filesystem, 'location', mock_location), mock.patch.object(builder.shell, 'run') as mock_run:
        builder.build('some_path', 'my_image', 'mytag')
        mock_run.assert_called_once()

        mock_run.assert_called_with(['docker', 'build', '-t', 'my_image:mytag', '.'], silent=True)


def test_builder_build_sets_default_tag():
    with mock.patch.object(builder.filesystem, 'location', mock_location), mock.patch.object(builder.shell, 'run') as mock_run:
        builder.build('some_path', 'my_image')
        mock_run.assert_called_once()

        mock_run.assert_called_with(['docker', 'build', '-t', 'my_image:latest', '.'], silent=True)


def test_builder_push_calls_correct_command():
    with mock.patch.object(builder.shell, 'run') as mock_run:
        builder.push('someuser/some_image', 'some_tag')
        mock_run.assert_called_once()

        mock_run.assert_called_with(['docker', 'push', 'someuser/some_image:some_tag'], silent=False)
